#--------------------------------------------------------------------------------------------
# ■マルチスレッド例
#
#   表示衝突制御はセマフォー操作、サブスレッドの終了検出は実行中スレッドの数で判別します。
#
#   論理的な手法については
#
#        拙著Kindle版「C言語によるマルチスレッドプログラミング」
#
#   を参照してください。
#
# -*- coding: UTF-8 -*-
import threading, time, datetime

numSub=0  # 実行中サブスレッドの数
S=1       # 出力箇所は 1 箇所なのでセマフォの値は 1 とする
#■セマフォの P 操作
def P():            # 繁忙待ちを避けるには pass の替わりに time.sleep(0.001)を入れる。
    global S        # パソコンでは繁忙待ちでも支障がないので pass とする。
    while S<=0:pass # UNIX等のように複数ユーザ使用が前提の場合は sleep()を使うべき。
    S-=1            
#■セマフォの V 操作    
def V():
    global S
    S+=1
#■セマフォのP/V操作による表示制御    
def printB(STR): #  P操作とV操作を外すと同時刻の表示が乱れることを確認してみよう
    P()
    print(STR)
    V()
#■サブスレッドの実装    
class subThread(threading.Thread): # サブスレッドの実装
    def __init__(self,no, nm, tm):
        super(subThread, self).__init__()
        self.no=no; self.nm = nm; self.tm = tm
        self.space=" "*20*self.no  # サブスレッドの表示位置を空白の数で表現する
    def run(self):
        global numSub
        printB(self.space+" *Start Sub %d* " % self.no) # サブスレッド開始
        st="Sub %d : " % self.no + datetime.datetime.today().strftime('%I:%M:%S')
        printB(self.space+ st)
        for i in range(self.nm): # 時間間隔 tm で nm 回実行
            time.sleep(self.tm)
            st="Sub %d : " % self.no + datetime.datetime.today().strftime('%I:%M:%S')
            printB(self.space+ st)
        printB(self.space+ " *End Sub %d* " % self.no)　# サブスレッド開始
        numSub -= 1 # サブスレッドを終わるとき実行中サブスレッドの個数を減ずる
#■メイン処理
if __name__ == '__main__':
    print()
    numSub=2 # 実行中サブスレッドを2個とする
    th_s1 = subThread(1,10, 2); th_s2 = subThread(2,5, 5)
    th_s1.start()             ; th_s2.start()
    printB(" *Start Main* ")
    st="Main : "+ datetime.datetime.today().strftime('%I:%M:%S')
    printB(st)
    for i in range(5):
        time.sleep(8)   #実行中サブスレッドが0個のとき全サブスレッド完了とみなす
        if numSub==0: printB("*Sub End Detect"); numSub=2 # 検出済みとみなす
        st="Main : "+ datetime.datetime.today().strftime('%I:%M:%S')
        printB(st)
    printB( " *End Main*")
    print()
